import numpy as np
import pickle
import random
import os
import torch
import glob

from datetime import datetime

def redistribute_probs(tensor):
    if tensor.dim() == 1:
        tensor = tensor.unsqueeze(0)
    # Get the indices of the maximum values in each row
    max_indices = torch.argmax(tensor, dim=1)
    
    # Create a copy of the origina
    # l tensor to avoid modifying it in-place
    adjusted_tensor = tensor.clone()
    
    # Iterate over each row in the tensor
    for i in range(tensor.size(0)):
        max_index = max_indices[i]
        
        # Compute the total reduction to be distributed
        reduction_sum = 0
        
        # Reduce all non-max elements by 4/5
        for j in range(tensor.size(1)):
            if j != max_index:
                reduction = tensor[i, j] * 1 / 5
                reduction_sum += tensor[i, j] - reduction
                adjusted_tensor[i, j] = reduction
        
        # Add the total reduction to the max element
        adjusted_tensor[i, max_index] += reduction_sum
        
    return adjusted_tensor

class Log:
    def __init__(self, args):
        self.args = args
        self.err_logs = {'train': [], 'val': [], 'test': []}
        self.nums = {'train': 0, 'val': 0, 'test': 0}
        self.log_file = open(os.path.join(args.checkpoint_dir, 'log.txt'), 'w')
        self.log_file.write('{}\n'.format(str(self.args)))
        
    def save_info(self, outputs, labels, ids, indices, type):
        
        _, inds= outputs.topk(k=len(self.args.labels), dim=1)
        pred = inds[:, 0]
        cor_inds = torch.nonzero(torch.eq(pred, labels)).view(-1)
        err_inds = torch.nonzero(torch.ne(pred, labels)).view(-1)
        
        if len(err_inds) != 0:
            self.nums[type] += len(err_inds)
            err_ids = ids[err_inds]
            err_labels = labels[err_inds]
            self.err_logs[type].append(torch.cat([err_ids.unsqueeze(1), err_labels.unsqueeze(1)], dim=1))
    
    def save_log(self, epoch, seed=None, test=False):
        for k in self.err_logs.keys():
            if self.nums[k] != 0:
                n_err = len(self.err_logs[k])
                n_cor = self.nums[k] - n_err
                acc = 1 - (n_err / self.nums[k])
                err_logs = torch.vstack(self.err_logs[k]).cpu().numpy()
                uniques, counts = np.unique(err_logs, axis=0, return_counts=True)
                self.err_logs[k] = np.concatenate((uniques, np.expand_dims(counts, axis=1)), axis=1)
                acc_info = '{} accuracy: {} / {} = {}'.format(k, n_cor, self.nums[k], acc)
                if test == False:
                    acc_info = 'Epoch [{}/{}] '.format(epoch+1, self.args.n_epochs) + acc_info
                if seed != None:
                    acc_info = 'Seed = {}\n'.format(seed) + acc_info
                self.log_file.write('{}\n'.format(acc_info))
                self.log_file.write('{} errors: {}\n\n\n'.format(k, self.err_logs[k]))
        self.err_logs = {'train': [], 'val': [], 'test': []}
        self.nums = {'train': 0, 'val': 0, 'test': 0}
    
def flatten_list(lst: list):
    return [item for sublist in lst for item in (flatten_list(sublist) if isinstance(sublist, list) else [sublist])]

def center_data(x: np.ndarray):
    return (x - x.mean(axis=0)) / x.std(axis=0)

def preprocess_args(args):
    args.seeds = [int(s) for s in args.seeds.split('+')]
    if args.log_path is None:
        args.log_path = args.checkpoint_path
    args.timestamp = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
    args.checkpoint_dir = os.path.join('./checkpoint', '{}_{}_{}'.format(args.model, args.seeds[0], args.data))
    if os.path.exists(args.checkpoint_dir) == False:
        os.mkdir(args.checkpoint_dir)
    args.data_dir = os.path.join('./dataset', args.data)

    # labels
    if args.data == 'BabyChillanto':
        args.labels = ['asphyxia', 'deaf', 'hunger', 'normal', 'pain']
    elif args.data == 'DonateACry':
        args.labels = ['belly_pain', 'burping', 'discomfort', 'hungry', 'tired']
        args.n_gender_classes = 2
        args.n_age_classes = 5
    elif args.data == 'Mix':
        args.labels = ['pain', 'hungry', 'asphyxia', 'deaf']
        args.remove_ids = [253, 251]
    else:
        raise ValueError('Unknow dataset {}'.format(args.data))
    if len(args.add_labels) != 0:
        args.labels += [label for label in args.add_labels.split('+')]
    
def fix_seed(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed) # if you are using multi-GPU.
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    
    
def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
    return res
    

def adjust_learning_rate(optimizer, epoch, base_lr, ajust_period=20):
    """Sets the learning rate to the initial LR decayed by 10 every 100 epochs"""
    lr = base_lr * (0.1 ** (epoch // ajust_period))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr



class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)

def rename_donateacry(data_dir, tgt_dir=None, baby_id=-1):
    '''
        baby_id: -1 if rename from scratch. If adding new samples to existing dataset, set baby_id as the last id of the existing one.
    '''
    if tgt_dir == None:
        tgt_dir = data_dir
    record2id = {}
    for dir in os.listdir(data_dir):
        cls_path = os.path.join(data_dir, dir)
        if tgt_dir != data_dir:
            os.makedirs(os.path.join(tgt_dir, dir))
        if os.path.isdir(cls_path):
            for file in os.listdir(cls_path):
                if file.lower().endswith(".wav"):
                    if dir.startswith('Full_'):
                        # Baby chillanto
                        record = file[0:2]
                        age = 'Null'
                        gender = 'Null'
                    else:
                        # DonateACry
                        infos = file.split('-')
                        age = infos[-2]
                        gender = infos[-3]
                        record = '-'.join(infos[0:4] + [gender, age])
                        
                    if record in record2id:
                        record2id[record][1] += 1
                        assigned_id, idx = record2id[record]
                    else:
                        baby_id += 1
                        assigned_id = baby_id
                        idx = 0
                        record2id[record] = [assigned_id, idx]

                    new_file = '-'.join([str(assigned_id), str(idx), gender, age]) + '.wav'
                    src_path = os.path.join(data_dir, dir, file)
                    tgt_path = os.path.join(tgt_dir, dir, new_file)
                    os.rename(src_path, tgt_path)
                    print('Rename {} to {}.'.format(src_path, tgt_path))
                    
if __name__ == '__main__':
    rename_donateacry('/home/minghao.fu/workspace/Letting-baby-talk-to-you/dataset/rename', baby_id=310)

                        


import torch
import torch.nn as nn
import torch.nn.functional as F

    
class mlp_attention(nn.Module):
    def __init__(self, in_feats, mid_feats, out_feats, n_classes):
        super(mlp_attention, self).__init__()
        self.fc1 = nn.Linear(in_feats, mid_feats)
        self.att1 = nn.Linear(in_feats, mid_feats)
        self.fc2 = nn.Linear(mid_feats, out_feats)
        self.att2 = nn.Linear(mid_feats, out_feats)
        self.fc3 = nn.Linear(out_feats, n_classes)
        
        self.softmax = nn.Softmax(dim=1)
        self.relu = nn.ReLU(inplace=True)
        
        self.fc4 = nn.Linear(out_feats, out_feats)
        self.fc5 = nn.Linear(out_feats, out_feats)
        

    def forward(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        x1 = self.fc1(x)
        att1 = self.softmax(self.att1(x))
        x1 = self.relu(x1 * att1)
        x2 = self.fc2(x1) 
        att2 = self.softmax(self.att2(x1))
        x2 = self.relu(x2 * att2)
        x3 = self.fc3(x2)

        return x3
    
    def forward_constrative(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        x1 = self.fc1(x)
        att1 = self.softmax(self.att1(x))
        x1 = self.relu(x1 * att1)
        x2 = self.fc2(x1) 
        att2 = self.softmax(self.att2(x1))
        x2 = self.relu(x2 * att2)
        x2 = self.fc4(x2)
        x2 = F.relu(x2)
        x2 = self.fc5(x2)
        # Normalize to unit hypersphere
        x2 = F.normalize(x2, dim=1)

        return x2




import torch
import torch.nn as nn

class resnet(nn.Module):
    def __init__(self, n_labels=4):
        super(resnet, self).__init__()
        self.conv1 = nn.Conv2d(1, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        self.layer2 = nn.Sequential(
            nn.Conv2d(64, 128, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
        )
        self.layer3 = nn.Sequential(
            nn.Conv2d(128, 256, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
        )
        self.layer4 = nn.Sequential(
            nn.Conv2d(256, 512, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace=True),
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, n_labels)

    def forward(self, x):
        x = x.unsqueeze(1) # add channel 1 as the new dimension
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x
    
class bin_resnet(nn.Module):
    def __init__(self):
        super(bin_resnet, self).__init__()
        self.conv1 = nn.Conv2d(1, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = nn.Sequential(
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
        )
        self.layer2 = nn.Sequential(
            nn.Conv2d(64, 128, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 128, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
        )
        self.layer3 = nn.Sequential(
            nn.Conv2d(128, 256, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
        )
        self.layer4 = nn.Sequential(
            nn.Conv2d(256, 512, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(512),
            nn.ReLU(inplace=True),
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, 2)

    def forward(self, x):
        x = x.unsqueeze(1) # add channel 1 as the new dimension
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        return x


from torch.utils.data import Dataset, DataLoader
import torch

class AudioDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels
    
    def __len__(self):
        return len(self.features)
    
    def __getitem__(self, idx):
        feature = torch.tensor(self.features[idx], dtype=torch.float32)
        label = torch.tensor(self.labels[idx], dtype=torch.long)
        return feature, label


import torch
import torch.nn as nn
import torch.nn.functional as F


def call_bn(bn, x):
    return bn(x)


class onedcnn(nn.Module):
    def __init__(self, n_classes=5, dropout=0.5):
        super(onedcnn, self).__init__()
        self.dropout_rate= dropout
        self.c1=nn.Conv2d(1,128,kernel_size=3,stride=1, padding=1)
        self.c2=nn.Conv2d(128,128,kernel_size=3,stride=1, padding=1)
        self.c3=nn.Conv2d(128,128,kernel_size=3,stride=1, padding=1)
        self.c4=nn.Conv2d(128,256,kernel_size=3,stride=1, padding=1)
        self.c5=nn.Conv2d(256,256,kernel_size=3,stride=1, padding=1)
        self.c6=nn.Conv2d(256,512,kernel_size=3,stride=1, padding=1)
        self.c7=nn.Conv2d(512,256,kernel_size=3,stride=1, padding=0)
        self.c8=nn.Conv2d(256,128,kernel_size=3,stride=1, padding=0)
        self.c9=nn.Conv2d(256,128,kernel_size=3,stride=1, padding=0)
        self.l_c1=nn.Linear(128,n_classes)

        self.fc1= nn.Linear(128, 128)
        self.fc2 = nn.Linear(128, 128)
        self.bn1=nn.BatchNorm2d(128)
        self.bn2=nn.BatchNorm2d(128)
        self.bn3=nn.BatchNorm2d(128)
        self.bn4=nn.BatchNorm2d(256)
        self.bn5=nn.BatchNorm2d(256)
        self.bn6=nn.BatchNorm2d(512)
        self.bn7=nn.BatchNorm2d(256)
        self.bn8=nn.BatchNorm2d(128)
        self.bn9=nn.BatchNorm2d(128)

    def forward(self, x):
        # print(x.shape)
        x = torch.unsqueeze(x, 1)
        # print(x.shape)
        h=x
        # print(x.shape)
        h=self.c1(h)
        # print(x.shape)
        h=F.leaky_relu(call_bn(self.bn1, h), negative_slope=0.01)
        h=self.c2(h)
        h=F.leaky_relu(call_bn(self.bn2, h), negative_slope=0.01)
        h=self.c3(h)
        h=F.leaky_relu(call_bn(self.bn3, h), negative_slope=0.01)
        h=F.max_pool2d(h, kernel_size=2, stride=2)
        h=F.dropout2d(h, p=self.dropout_rate)

        h=self.c4(h)
        h=F.leaky_relu(call_bn(self.bn4, h), negative_slope=0.01)
        h=self.c5(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn5, h), negative_slope=0.01)
        h=self.c6(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn6, h), negative_slope=0.01)
        h=F.max_pool2d(h, kernel_size=2, stride=2)
        h=F.dropout2d(h, p=self.dropout_rate)
        # print(h.shape)
        h=self.c7(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn7, h), negative_slope=0.01)
        h=self.c8(h)
        # print("c8",h.shape)
        h=F.leaky_relu(call_bn(self.bn8, h), negative_slope=0.01)
        # print("c8",h.shape)
        # h=self.c9(h)
        # # print("c9", h.shape)
        # h=F.leaky_relu(call_bn(self.bn9, h), negative_slope=0.01)
        h=F.avg_pool2d(h, kernel_size=(h.data.shape[2],1))
        
        # print(h.shape)
        h = h.view(h.size(0), h.size(1))
        # print(h.shape)
        logit=self.l_c1(h)

        return logit



    def forward_contrastive(self, x):
        # print(x.shape)
        x = torch.unsqueeze(x, 1)
        # print(x.shape)
        h=x
        # print(x.shape)
        h=self.c1(h)
        # print(x.shape)
        h=F.leaky_relu(call_bn(self.bn1, h), negative_slope=0.01)
        h=self.c2(h)
        h=F.leaky_relu(call_bn(self.bn2, h), negative_slope=0.01)
        h=self.c3(h)
        h=F.leaky_relu(call_bn(self.bn3, h), negative_slope=0.01)
        h=F.max_pool2d(h, kernel_size=2, stride=2)
        h=F.dropout2d(h, p=self.dropout_rate)

        h=self.c4(h)
        h=F.leaky_relu(call_bn(self.bn4, h), negative_slope=0.01)
        h=self.c5(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn5, h), negative_slope=0.01)
        h=self.c6(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn6, h), negative_slope=0.01)
        h=F.max_pool2d(h, kernel_size=2, stride=2)
        h=F.dropout2d(h, p=self.dropout_rate)
        # print(h.shape)
        h=self.c7(h)
        # print(h.shape)
        h=F.leaky_relu(call_bn(self.bn7, h), negative_slope=0.01)
        h=self.c8(h)
        # print("c8",h.shape)
        h=F.leaky_relu(call_bn(self.bn8, h), negative_slope=0.01)
        # print("c8",h.shape)
        # h=self.c9(h)
        # # print("c9", h.shape)
        # h=F.leaky_relu(call_bn(self.bn9, h), negative_slope=0.01)
        h=F.avg_pool2d(h, kernel_size=(h.data.shape[2],1))
        
        # print(h.shape)
        h = h.view(h.size(0), h.size(1))
        h=self.fc1(h)
        h=F.leaky_relu(h, negative_slope=0.01)
        h=self.fc2(h)
        h = F.normalize(h, dim=1)

        return h
   



import os
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import matplotlib
import matplotlib.pyplot as plt

from importlib import import_module

class Loss(nn.modules.loss._Loss):
    def __init__(self, args):
        super(Loss, self).__init__()
        self.loss = []
        self.epoch = 0
        self.log_path = os.path.join(args.checkpoint_dir)
        self.loss_module = nn.ModuleList()
        self.log = {'val_accuracy': []}
        for loss in args.loss.split('+'):
            weight, loss_type = loss.split('*')
            loss_module = import_module('loss.' + loss_type)
            loss_function = getattr(loss_module, loss_type)(args)
            
            self.log['train_' + loss_type] = []
            self.log['val_' + loss_type] = []

            self.loss.append({
                'type': loss_type,
                'weight': float(weight),
                'function': loss_function
            })

    def forward(self, projections, targets):
        loss_sum = 0
        for i, l in enumerate(self.loss):
            l['loss'] = l['function'](projections, targets)
            loss_sum += l['weight'] * l['loss']
            self.loss[i] = l
        return loss_sum
    
    def save_log(self, train_loss, val_loss, val_accuracy):
        self.epoch += 1
        for i, l in enumerate(self.loss):
            self.log['train_' + l['type']].append(train_loss)
            self.log['val_' + l['type']].append(val_loss)
        self.log['val_accuracy'].append(val_accuracy)

    def step(self):
        for l in self.get_loss_module():
            if hasattr(l, 'scheduler'):
                l.scheduler.step()

    def plot_loss(self, name='Log'):
        axis = np.linspace(1, self.epoch, self.epoch)
        fig, (loss_ax, acc_ax) = plt.subplots(nrows=2, sharex=True)

        plt.title(name)
        plt.grid(True)
        plt.xlabel('Epochs')
        loss_ax.set_ylabel('Loss')
        acc_ax.set_ylabel('Accuracy')
        for label, loss_value in self.log.items():
            if label == 'val_accuracy':
                acc_ax.plot(axis, loss_value, label=label)
            else:
                loss_ax.plot(axis, loss_value, label=label)

        # Add a legend to the plot
        lines1, labels1 = loss_ax.get_legend_handles_labels()
        lines2, labels2 = acc_ax.get_legend_handles_labels()
        fig.legend(lines1 + lines2, labels1 + labels2, loc='upper right')

        plt.savefig(os.path.join(self.log_path, name + '.pdf'))
        plt.close(fig)
    
    def print_loss_coomponent(self):
        print('')



import torch 
import argparse
import torch
import torch.nn as nn
import torch.optim as optim
import os
import torch.nn.functional as F

from tqdm import tqdm
from torch.utils.data import Dataset, DataLoader, random_split
from sklearn.manifold import TSNE
from importlib import import_module

from loss import Loss
from utils import fix_seed, preprocess_args, AverageMeter, accuracy, Log
    
def train_babychillanto(args, model, dataset, log, seed):
    
    criterion = Loss(args)
    optimizer = optim.Adam(model.parameters(), lr=args.lr)

    train_loader = DataLoader(dataset.train_dataset, batch_size=args.batch_size, shuffle=True, drop_last=True)
    val_loader = DataLoader(dataset.val_dataset, batch_size=args.batch_size, shuffle=False)
    test_loader = DataLoader(dataset.test_dataset, batch_size=args.batch_size, shuffle=False)

    best_model = None
    best_val_performance = float('-inf') 
    for epoch in range(args.n_epochs):
        n_top1 = AverageMeter('Acc@1', ':6.2f')
        train_loss = AverageMeter('Acc@1', ':6.2f')
        n_top2 = AverageMeter('Acc@1', ':6.2f')
        
        cls_train_top1= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        cls_train_top2= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        cls_val_top1= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        cls_val_top2= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        cls_test_top1= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        cls_test_top2= [AverageMeter('Acc@1', ':6.2f') for _ in range(len(args.labels))]
        
        for batch_data in train_loader:
            b_audios, b_labels, b_ids, b_indices = batch_data
            optimizer.zero_grad()
            outputs = model(b_audios)
            
            _, indices= outputs.topk(k=len(args.labels), dim=1)
            pred = indices[:, 0].squeeze(0)
            log.save_info(outputs, b_labels, b_ids, b_indices, 'train')
                
            loss = criterion(outputs, b_labels)
            loss.backward()
            optimizer.step()
            n_acc1,n_acc2 = accuracy(outputs, b_labels, topk=(1, 2))
            n_top1.update(n_acc1.item(), b_audios.size(0))
            n_top2.update(n_acc2.item(), b_audios.size(0))
            train_loss.update(loss.item(), 1)
            # print(loss.item())
            
            for i, cls in enumerate(args.labels):
                _inds = torch.nonzero(torch.eq(b_labels, i)).view(-1)
                if len(_inds) > 0:
                    cls_acc1,cls_acc2 = accuracy(outputs[_inds], b_labels[_inds], topk=(1, 2))
                    cls_train_top1[i].update(cls_acc1.item(), b_audios.size(0))
                    cls_train_top2[i].update(cls_acc2.item(), b_audios.size(0))
        
        model.eval()
        val_loss = 0
        correct = 0
        id2wrongCount = {id : 0 for id in dataset.val_dataset.data['id'].tolist()}
        cls2wrongCount = {cls : 0 for cls in args.labels}
        
        with torch.no_grad():
            val_top1= AverageMeter('Acc@1', ':6.2f')
            val_top2= AverageMeter('Acc@1', ':6.2f')
            val_loss = AverageMeter('Acc@1', ':6.2f')
            
            for vb_audios, vb_labels, vb_ids, vb_indices in val_loader:
                # Compute predictions and loss
                voutputs = model(vb_audios)
                log.save_info(voutputs, vb_labels, vb_ids, vb_indices, 'val')
                loss = criterion(voutputs, vb_labels)
                
                _, indices= voutputs.topk(k=len(args.labels), dim=1)
                pred = indices[:, 0].squeeze(0)
                
                correct += pred.eq(vb_labels).sum().item()
    
                correct_indices = torch.nonzero(torch.eq(pred, vb_labels)).view(-1)
                mistake_indices = torch.nonzero(torch.ne(pred, vb_labels)).view(-1)
                if len(mistake_indices) != 0:
                    out_types = [[args.labels[value] for value in row] for row in indices[mistake_indices].tolist()]
                    y_type = [args.labels[value] for value in vb_labels[mistake_indices].tolist()]

                    wrong_ids = vb_ids[mistake_indices].tolist()
                    for id in wrong_ids:
                        id2wrongCount[id] += 1
                    wrong_cls = [args.labels[i] for i in vb_labels[mistake_indices].tolist()]
                    for cls in wrong_cls:
                        cls2wrongCount[cls] += 1
                        
                for i, cls in enumerate(args.labels):
                    _inds = torch.nonzero(torch.eq(vb_labels, i)).view(-1)
                    if len(_inds) > 0:
                        cls_acc1,cls_acc2 = accuracy(voutputs[_inds], vb_labels[_inds], topk=(1, 2))
                        cls_val_top1[i].update(cls_acc1.item(), vb_audios.size(0))
                        cls_val_top2[i].update(cls_acc2.item(), vb_audios.size(0))
                    
                val_acc1,val_acc2 = accuracy(voutputs, vb_labels, topk=(1, 2))
                val_top1.update(val_acc1.item(), vb_audios.size(0))
                val_top2.update(val_acc2.item(), vb_audios.size(0))
                val_loss.update(loss.item(), 1)
                
        for id, count in id2wrongCount.items():
            id2wrongCount[id] = count / dataset.val_data['id'].eq(id).sum().item() 
            
        for cls, count in cls2wrongCount.items():
            cls_num = dataset.val_data['label'].eq(args.labels.index(cls)).sum().item()
            cls_accuracy = round(count / cls_num, 3)
            cls2wrongCount[cls] = '{} / {} = {}'.format(count, cls_num, cls_accuracy)
        
        id2wrongCount = {k: round(v, 3) for k, v in id2wrongCount.items() if v > 0.3}
        avg_train_acc_top1 = sum([i.avg for i in cls_train_top1]) / len(args.labels)
        avg_train_acc_top2 = sum([i.avg for i in cls_train_top2]) / len(args.labels)
        avg_val_acc_top1 = sum([i.avg for i in cls_val_top1]) / len(args.labels)
        avg_val_acc_top2 = sum([i.avg for i in cls_val_top2]) / len(args.labels)
        if avg_val_acc_top1 > best_val_performance:
            log.save_log(epoch, seed=seed)
            best_val_performance = avg_val_acc_top1
            best_model = model.state_dict() 
            torch.save(best_model, os.path.join(args.checkpoint_dir, 'best_model.pth'))
            
            print(f"Epoch [{epoch+1}/{args.n_epochs}]: Training Loss = {train_loss.avg: .3f}, Train Accuracy Top1/2 = {avg_train_acc_top1: .3f}/{avg_train_acc_top2: .3f}, \
                Val Accuracy Top1/2 = {avg_val_acc_top1: .3f} / {avg_val_acc_top2: .3f}")
            #print('Val: Error rate in each class: {}'.format(cls2wrongCount))
            cls_val_acc_info = 'Validation accuracy top1/2 in each class: ' 
            for i in range(len(args.labels)):
                cls_val_acc_info += '{} : {:.3f} / {:.3f} '.format(args.labels[i], cls_val_top1[i].avg, cls_val_top2[i].avg)
            cls_train_acc_info = 'Training accuracy top1/2 in each class: ' 
            for i in range(len(args.labels)):
                cls_train_acc_info += '{} : {:.3f} / {:.3f} '.format(args.labels[i], cls_train_top1[i].avg, cls_train_top2[i].avg)
            print(cls_val_acc_info)
            print(cls_train_acc_info)
            
        criterion.save_log(train_loss.avg, val_loss.avg, val_top1.avg)
        
        #print(f"Epoch [{epoch+1}/{args.n_epochs}]: Training loss = {train_loss.avg: .3f}, train Accuracy = {n_top1.avg: .3f}/{n_top2.avg: .3f} \
              #Val Accuracy = {val_acc1.item(): .3f} Val Accuracy = {val_top1.avg: .3f}/{val_top2.avg: .3f}, Best Val Accuracy = {best_val_performance: .3f}")
    criterion.plot_loss()
    
    model.load_state_dict(best_model)
    model.eval()
    correct = 0
    inf_times = []
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    with torch.no_grad():
        test_top1= AverageMeter('Acc@1', ':6.2f')
        test_top2= AverageMeter('Acc@1', ':6.2f')
        test_loss = AverageMeter('Acc@1', ':6.2f')
        for audios, labels, ids, indices in test_loader:
            start.record()
            outputs = model(audios)
            end.record()
            torch.cuda.synchronize()
            inf_times.append(start.elapsed_time(end))

            log.save_info(outputs, labels, ids, indices, 'test')
            loss = criterion(outputs, labels)
            _, predicted = torch.max(outputs, 1)
            probs = torch.nn.functional.softmax(outputs, dim=-1)
            print(f'Probability: \n {probs}')
            print(predicted == labels)
            
            test_acc1,test_acc2 = accuracy(outputs, labels, topk=(1, 2))
            test_top1.update(test_acc1.item(), audios.size(0))
            test_top2.update(test_acc2.item(), audios.size(0))
            test_loss.update(loss.item(), 1)
    
    log.save_log(epoch, test=True)
    # test_loss /= len(test_loader.dataset)
    # test_acc = 100 * correct / len(test_loader.dataset)
    
    print(f"Seed = {seed}: Test loss = {test_loss.avg: .3f}, test Accuracy = {test_top1.avg: .3f}/{test_top2.avg: .3f}, average inference time per {args.seg_len}s audio: {sum(inf_times) / len(inf_times): .3f}ms")
    
    return best_model, best_val_performance, test_top1.avg

# def train_donateacry(args, model, dataset):

#     criterion = Loss(args)
#     optimizer = optim.Adam(model.parameters(), lr=args.lr)

#     train_loader = DataLoader(dataset.train_dataset, batch_size=args.batch_size, shuffle=True, drop_last=True)
#     val_loader = DataLoader(dataset.val_dataset, batch_size=args.batch_size, shuffle=False)
#     test_loader = DataLoader(dataset.test_dataset, batch_size=args.batch_size, shuffle=False)

#     best_model = None
#     best_val_performance = float('-inf') 
#     for epoch in range(args.n_epochs):
#         n_top1 = AverageMeter('Acc@1', ':6.2f')
#         train_loss = AverageMeter('Acc@1', ':6.2f')
#         n_top2 = AverageMeter('Acc@1', ':6.2f')
#         for b_audios, b_labels, b_ids, b_genders, b_ages in train_loader:
#             optimizer.zero_grad()
#             outputs = model(b_audios)#, b_genders, b_ages)
#             loss = criterion(outputs, b_labels)
#             loss.backward()
#             optimizer.step()
#             n_acc1,n_acc2 = accuracy(outputs, b_labels, topk=(1, 2))
#             n_top1.update(n_acc1.item(), b_audios.size(0))
#             n_top2.update(n_acc2.item(), b_audios.size(0))
#             train_loss.update(loss.item(), 1)
#             # print(loss.item())

#         model.eval()
#         val_loss = 0
#         correct = 0
#         id2wrongCount = {id : 0 for id in dataset.val_dataset.data['id'].tolist()}
#         wrong_ids = []
#         with torch.no_grad():
#             val_top1= AverageMeter('Acc@1', ':6.2f')
#             val_top2= AverageMeter('Acc@1', ':6.2f')
#             val_loss = AverageMeter('Acc@1', ':6.2f')
#             for vb_audios, vb_labels, vb_ids, vb_genders, vb_ages in val_loader:
#                 # Compute predictions and loss
#                 voutputs = model(vb_audios)#, vb_genders, vb_ages)
#                 loss = criterion(voutputs, vb_labels)     
#                 pred = voutputs.argmax(dim=1, keepdim=True)
#                 correct += pred.squeeze(1).eq(vb_labels).sum().item()
#                 correct_indices = torch.nonzero(pred.squeeze(1).eq(vb_labels)).squeeze()
#                 mistake_indices = torch.nonzero(torch.ne(pred.squeeze(1), vb_labels)).squeeze()
#                 #wrong_ids = [vb_ids[mistake_indices]] if len(mistake_indices) == 1 else vb_ids[mistake_indices].tolist()
#                 if len(mistake_indices.shape) != 0:
#                     wrong_ids.extend(vb_ids[mistake_indices].tolist())
                    
#                 for id in wrong_ids:
#                     id2wrongCount[id] += 1
#                 val_acc1,val_acc2 = accuracy(voutputs, vb_labels, topk=(1, 2))
#                 val_top1.update(val_acc1.item(), b_audios.size(0))
#                 val_top2.update(val_acc2.item(), b_audios.size(0))
#                 val_loss.update(loss.item(), 1)


#         # Check if current model has better performance than previous best model
#         if val_top1.avg > best_val_performance:
#             best_val_performance = val_top1.avg
#             best_model = model.state_dict()  # Save the model's state dictionary
#             torch.save(best_model, os.path.join(args.checkpoint_dir, 'best_model.pth'))
        
#         criterion.save_log(train_loss.avg, val_loss.avg, val_top1.avg)
        
#         print(f"Epoch [{epoch+1}/{args.n_epochs}]: Training loss = {train_loss.avg: .3f}, train Accuracy = {n_top1.avg: .3f}/{n_top2.avg: .3f} \
#               Val Accuracy = {val_top1.avg: .3f}/{val_top2.avg: .3f}, Best Val Accuracy = {best_val_performance: .3f}")
#     criterion.plot_loss()
#     return best_model, best_val_performance


import argparse
import torch
import torch.nn as nn
import torch.optim as optim
import os
from torch.nn import functional as F
import numpy as np

from loss import Loss

from tqdm import tqdm
from torch.utils.data import Dataset, DataLoader, random_split
from sklearn.manifold import TSNE
from importlib import import_module

from utils import fix_seed, preprocess_args, AverageMeter, accuracy, Log
from trainer import train_babychillanto
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Descriptions')

    parser.add_argument('--data', type=str, default='BabyChillanto', help='Select dataset')
    parser.add_argument('--checkpoint_path', type=str, default=None, help='Path to checkpoint file')  
    parser.add_argument('--log_path', type=str, default=None, help='Path to logging file')  
    parser.add_argument('--pattern', type=str, default=r'^\d+', help='Regex pattern of id extraction')
    parser.add_argument('--split_type', type=str, default='class', help='Type of data split')
    parser.add_argument('--n_mfcc_coeffs', type=int, default=20, help='Number of MFCC coefficients')
    parser.add_argument('--seg_len', type=int, default=4, help='Segment length of audio')
    parser.add_argument('--shift', type=float, default=1, help='Shift length of each segment')
    parser.add_argument('--val_rate', type=float, default=0.2, help='Validation data ratio')
    parser.add_argument('--n_dim',type=int, default=128, help='Transformer dimension')
    parser.add_argument('--save_path', type=str, default=None, help='Path to save output')
    parser.add_argument('--load_path', type=str, default=None, help='Path to load input')
    parser.add_argument('--seeds', type=str, default='42', help='Random seeds (for multiple training)')
    parser.add_argument('--seg_save_dir', type=str, default=None, help='Saving segmented audio files')
    parser.add_argument('--reweight', action='store_true', help='Reweighting')
    parser.add_argument('--add_labels', type=str, default='', help='Added labels')
    parser.add_argument('--center_data', type=bool, default=False, help='Center data for training or not')
    
    parser.add_argument('--loss', type=str, default='1*cross_entropy_loss+1*supervised_contrastive_loss', help='Loss function combination')
    parser.add_argument('--temperature', type=float, default=0.07, help='Temperature for contrastive loss function')
    parser.add_argument('--model', type=str, default='mlp', help='Model name')
    parser.add_argument('--dropout', type=float, default=None, help='Dropout')
    parser.add_argument('--lr', type=float, default=1e-4, help='Learning rate')
    parser.add_argument('--n_epochs', type=int, default=20, help='Number of epochs')
    parser.add_argument('--batch_size', type=int, default=8, help='Batch size')
    
    
    parser.add_argument('--gender_embed_dim', type=int, default=512, help='Embedding dim of gender') 
    parser.add_argument('--age_embed_dim', type=int, default=512, help='Embedding dim of age')
    
    args = parser.parse_args()
    

preprocess_args(args)
log = Log(args)
val_accuracies = []
test_accuracies = []
for seed in args.seeds:
    fix_seed(seed)
    
    data_module = getattr(import_module('data.' + args.data), args.data)
    dataset = data_module(args)
    if args.reweight:
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        args.cls_weights = 1.0 / torch.Tensor(dataset.cls_num, device=device).float()
    
    module = getattr(import_module('models.' + args.model), args.model)
    args.in_feats = dataset.train_dataset.n_flatten_feats
    
    if args.model == 'mlp':
        model = module(args.in_feats, 2000, 64, len(args.labels), args.dropout)
    elif args.model == 'mlp_attention':
        model = module(args.in_feats, 64, 64, len(args.labels))
    elif args.model == 'Transformer_model':
        model = module(args.in_feats, args.n_dim, len(args.labels))
    else:
        model = module(args)
        
    if torch.cuda.is_available():
        model = model.cuda()
    
    if args.checkpoint_path is not None:
        model.load_state_dict(torch.load(args.checkpoint_path))
    
    trainer = getattr(import_module('train'), 'train_babychillanto')
    best_model, best_val_performance, best_test_performance = trainer(args, model, dataset, log, seed)
    val_accuracies.append(best_val_performance)
    test_accuracies.append(best_test_performance)
    
print('Val accuracies: {}, mean: {}, var: {}'.format(val_accuracies, np.mean(val_accuracies), np.var(val_accuracies)))
print('Test accuracies: {}, mean: {}, var: {}'.format(test_accuracies, np.mean(test_accuracies), np.var(test_accuracies)))


from flask import Flask, request, jsonify
import torch
import librosa
import numpy as np
import random

app = Flask(__name__)

@app.route('/upload', methods=['POST'])
def read_audio_to_tensor(sample_method='auto_detect', 
                         seg_len=2, 
                         n_mfcc_coeffs=20, 
                         device=None, # if GPUs
                         ):
    if sample_method not in ['random', 'first', 'auto_detect']:
        raise ValueError("Sample method must be 'random', 'first', or 'auto_detect'")
    
    if 'file' not in request.files:
        return jsonify(error='No file part')

    file = request.files['file']

    if file.filename == '':
        return jsonify(error='No selected file')

    # Assume the file received ends with .wav
    if file and file.filename.endswith('.wav'):
        if sample_method == 'first':
            seg, sr = librosa.load(file, duration=seg_len)
            # Add white noise
            # if id in train_ids:
            #     noise_factor = 0.0001
            #     white_noise = np.random.randn(len(y)) * noise_factor
            #     y = y + white_noise
        else:
            y, sr = librosa.load(file)
            seg_len_size = seg_len * sr
            if seg_len_size > len(y):
                y = np.pad(y, (0, seg_len_size - len(y)), mode='constant')
            segments = [y[i : i + seg_len_size] for i in range(0, len(y), seg_len_size)] # shift window runs too slow
            if sample_method == 'random':
                seg = random.choice(segments)
            elif sample_method == 'auto_detect':
                energies = [np.sum(segment**2) for segment in segments]
                seg = segments[np.argmax(energies)]

        mfcc = librosa.feature.mfcc(y=seg, sr=sr, n_mfcc=n_mfcc_coeffs).T

        #n_frames = mfcc.shape[0]
        pitch = librosa.yin(seg, fmin=75, fmax=600)
        #intensity = librosa.feature.rms(y=y).flatten()
        feature = np.concatenate((mfcc, np.expand_dims(pitch, axis=1)), axis=1)
        
        feature_tensor = torch.tensor(feature).unsqueeze(0) # As model input, shape: (1, n_frames, n_mfcc_coeffs + 1)
        if device is not None:
            feature_tensor = feature_tensor.to(device)

        # You can now work with the tensor as needed
        # ...
        # ...
        # ...

        return jsonify(success=True)

    return jsonify(error='Invalid file type')

if __name__ == '__main__':
    app.run(debug=True)



import torch
import torch.nn as nn

class Transformer_model(nn.Module):
    def __init__(self, in_dim, n_dim, n_classes):
        super(Transformer_model, self).__init__()
        self.fc = nn.Linear(in_dim, n_dim)
        encoder_layer = nn.TransformerEncoderLayer(d_model=n_dim, nhead=8)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=6)
        self.classifier = nn.Linear(n_dim, n_classes)
        
    def forward(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        h = self.fc(x)
        y = self.transformer_encoder(h)
        out = self.classifier(y)
        return out


import os
import torch
import torch.nn as nn
import torchaudio
import re
import librosa
import pickle
import numpy as np
import random
import glob

from torch.utils.data import Dataset, DataLoader, random_split
from tqdm import tqdm

from utils import center_data, flatten_list

class SegmentedAudio(Dataset):
    def __init__(self, data):
        self.data = data
        self.len = data['audio'].shape[0]
        if len(data['audio'].shape) == 3:
            self.n_flatten_feats = data['audio'].shape[1] * data['audio'].shape[2]
        elif len(data['audio'].shape) == 2:
            self.n_flatten_feats = data['audio'].shape[-1]
        else:
            raise ValueError('Feature shape {} is not available.'.format(data['audio'].shape))
            
    def __len__(self):
        return self.len

    def __getitem__(self, idx):
        return self.data['audio'][idx], self.data['label'][idx], self.data['id'][idx], self.data['gender'][idx], self.data['age'][idx]
    
    def shape(self):
        return self.data['audio'].shape


class DonateACry:
    def __init__(self, args):
        self.data_dir = args.data_dir
        self.labels = args.labels
        self.seg_len = args.seg_len
        self.shift = args.shift
        self.n_mfcc_coeffs = args.n_mfcc_coeffs
        self.save_path = args.save_path
        self.load_path = args.load_path
        self.val_rate = args.val_rate
        self.split_type = args.split_type
        self.ages = ['04', '48', '26', '72', '22']
        self.id2audios, self.n_baby, self.baby_ids = self.get_dataset_ids()
        self.train_ids, self.val_ids = self.generate_train_val_ids(self.baby_ids, self.n_baby, self.val_rate, self.split_type, self.load_path, self.save_path)

        if self.seg_len == 1:
            train_data, val_data = self.load_one_second_audio(self.train_ids, self.val_ids, self.n_mfcc_coeffs)
        elif self.seg_len == 0:
            # 4 feature
            train_data, val_data = self.load_full_audio(self.train_ids, self.val_ids, self.n_mfcc_coeffs)
        else:
            train_data, val_data = self.load_n_seconds_audio_shift(self.seg_len, self.shift, self.train_ids, self.val_ids, self.n_mfcc_coeffs)
                
        if torch.cuda.is_available():
            for data in [train_data, val_data]:
                for k in data.keys():
                    data[k] = data[k].cuda()
        
        all_data = {'audio': [], 'label': [], 'id': []}
        all_data['audio'] = torch.cat([train_data['audio'], val_data['audio']], axis=0)
        all_data['label'] = torch.cat([train_data['label'], val_data['label']], axis=0)
        all_data['id'] = torch.cat([train_data['id'], val_data['id']], axis=0)

        self.all_data = all_data
        self.train_data = train_data
        self.val_data = val_data
        
        self.train_dataset = SegmentedAudio(train_data)
        self.val_dataset = SegmentedAudio(val_data)
        self.all_dataset = SegmentedAudio(all_data)
            
        print('Training samples: {}, validation samples: {}, total samples: {}'.format(len(self.train_dataset), \
            len(self.val_dataset), len(self.all_dataset)))
        
    def get_dataset_ids(self):
        audio_dirs = [os.path.join(self.data_dir, label) for label in self.labels]
        
        id2audios = {}
        baby_ids = []
        n_baby = 0
        for idx, dir in enumerate(audio_dirs):
            class_ids = []
            for file in os.listdir(dir):
                if file.lower().endswith(".wav"):
                    infos = file.split('-')
                    assert len(infos) == 3, "{} format is not true.".format(file)
                    id = int(infos[0])
                    if id in id2audios:
                        id2audios[id].append(file)
                    else:
                        id2audios[id] = [file]
                        class_ids.append(id)
                        n_baby += 1

            baby_ids.append(class_ids)
            
        print("--- Total baby number: {}, class distribution: {}".format(n_baby, [f"{self.labels[i]}: {len(cls)}" \
                for i, cls in enumerate(baby_ids)]))
        
        return id2audios, n_baby, baby_ids
        
    def generate_train_val_ids(self, individual_ids, n_baby, val_rate, split_type, save_path, load_path):
    
        if load_path is not None:
            with open(load_path, "rb") as file:
                data = pickle.load(file)
            train_ids = data['train']
            val_ids = data['val']
            assert len(val_ids + train_ids) == len(flatten_list(individual_ids))
            
        else:
            if split_type == 'class':
                train_ids = []
                val_ids = []
                for cls in individual_ids:
                    random.shuffle(cls)
                    n_val = int(len(cls) * val_rate)
                    val_ids.extend(cls[:n_val])
                    train_ids.extend(cls[n_val:])
                    
            elif split_type == 'baby':
                
                individual_ids = flatten_list(individual_ids)
                random.shuffle(individual_ids)
                n_val_individual = int(val_rate * n_baby)
                n_train_individuals = n_baby - n_val_individual
                train_ids = individual_ids[:n_train_individuals]
                val_ids = individual_ids[n_train_individuals:]
            
            else:
                raise ValueError('Unknow splitting type {}'.format(split_type))
            
            if save_path is not None:
                data = {'train': train_ids, 'val': val_ids}
                with open(save_path, "wb") as file:
                    pickle.dump(data, file)
        
        return train_ids, val_ids

    def load_full_audio(self, train_ids, val_ids, n_mfcc_coeffs, prefix='Full_'):
        train_data = {'audio': [], 'label': [], 'id': []}
        val_data = {'audio': [], 'label': [], 'id': []}
        full_audio_dirs = [self.data_dir + prefix + label for label in self.labels]
        print('--- Loading full audios')
        save_data=[]
        for i, dir in enumerate(full_audio_dirs):
            for file in os.listdir(dir):
                if file.lower().endswith(".wav"):
                    waveform, sample_rate = torchaudio.load(os.path.join(dir, file))
                    
                    sound_velocity = 343
                    wavelength = torch.tensor(sound_velocity / sample_rate).cuda()
                    
                    n_fft = 1024
                    win_length = None
                    hop_length = 512
                    n_mels = 256
                    n_mfcc = 256
                    spectrogram = T.Spectrogram(
                        n_fft=n_fft,
                        win_length=win_length,
                        hop_length=hop_length,
                        center=True,
                        pad_mode="reflect",
                        power=2.0,
                    )
                    spec = spectrogram(waveform)
                    
                    mfcc_transform = T.MFCC(
                        sample_rate=sample_rate,
                        n_mfcc=n_mfcc,
                        melkwargs={
                            "n_fft": n_fft,
                            "n_mels": n_mels,
                            "hop_length": hop_length,
                            "mel_scale": "htk",
                        },
                    )
                    mfcc = mfcc_transform(waveform).cuda()
                    pitch = F.detect_pitch_frequency(waveform, sample_rate).cuda()
                    
                    feature = torch.stack([mfcc.mean(), mfcc.var(), pitch.mean(), pitch.var(), wavelength], dim=0)
                    
                    id = int(re.findall(r'\d+', file)[0])
                    
                    feature_label_id = [mfcc.mean().cpu(), mfcc.var().cpu(), pitch.mean().cpu(), pitch.var().cpu(), wavelength.cpu(), i, id]
                    
                    save_data.append(feature_label_id)
                    
                    if id in train_ids:
                        train_data['audio'].append(feature)
                        train_data['label'].append(i)
                        train_data['id'].append(id)
                    elif id in val_ids:
                        val_data['audio'].append(feature)
                        val_data['label'].append(i)
                        val_data['id'].append(id)
                    else:
                        print('Data {}/{} is not included.'.format(dir, file))

            
        for data in [train_data, val_data]:
            data['audio'] =  torch.stack(data['audio'], dim=0)
            data['label'] = torch.tensor(data['label'])
            data['id'] = torch.tensor(data['id'])
            
        np.savetxt("./data.csv", np.array(save_data), delimiter=",")
        return train_data, val_data
        
    def load_one_second_audio(self, train_ids, val_ids, n_mfcc_coeffs, prefix='1s_'):
        data_idx = 0 
        train_data = {'audio': [], 'label': [], 'id': []}
        val_data = {'audio': [], 'label': [], 'id': []}
        one_second_audio_dirs = [self.data_dir + prefix + label for label in self.labels]
        print('--- Loading 1 second audios')
        for i, dir in enumerate(one_second_audio_dirs):
            for file in os.listdir(dir):
                if file.endswith(".wav"):
                    if data_idx == 1921: # dirty data
                        #print(dir + '/' + file)
                        data_idx += 1
                        continue
                    
                    y, sr = librosa.load(os.path.join(dir, file))
                    if sr != 22050:
                        print(sr, file)
                    
                    mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=n_mfcc_coeffs).T
                    #n_frames = mfcc.shape[0]
                    pitch = np.expand_dims(librosa.yin(y, fmin=75, fmax=600), axis=1)
                    #intensity = librosa.feature.rms(y=y).flatten()
                    feature = np.concatenate((mfcc, pitch), axis=1)
                    
                    id = int(file[2:4])
                    if id in train_ids:
                        train_data['audio'].append(feature)
                        train_data['label'].append(i)
                        train_data['id'].append(id)
                    elif id in val_ids:
                        val_data['audio'].append(feature)
                        val_data['label'].append(i)
                        val_data['id'].append(id)
                    else:
                        print('Data {}/{} is not included.'.format(dir, file))
                    
                    data_idx += 1
            
        for data in [train_data, val_data]:
            data['audio'] = torch.from_numpy(center_data(np.stack(data['audio']))).float()
            data['label'] = torch.tensor(data['label'])
            data['id'] = torch.tensor(data['id'])
            
        return train_data, val_data

    def load_n_seconds_audio(self, seg_len, train_ids, val_ids, n_mfcc_coeffs):
        train_data = {'audio': [], 'label': [], 'id': []}
        val_data = {'audio': [], 'label': [], 'id': []}
        full_audio_dirs = [os.path.join(self.data_dir, label) for label in self.labels]
        print('--- Loading and segmenting {} second audios'.format(seg_len))
        for i, dir in enumerate(full_audio_dirs):
            for file in os.listdir(dir):
                if file.endswith(".wav"):
                    
                    y, sr = librosa.load(os.path.join(dir, file))
                    seg_samples = int(seg_len * sr)
                    total_segs = len(y) // seg_samples
                    id = file.split('-')[0]
                    gender = file.split('-')[1]
                    age = file.split('-')[2]
                    for seg_index in range(total_segs):
                        start = seg_index * seg_samples
                        end = (seg_index + 1) * seg_samples
                        segment = y[start:end]
                        
                        mfcc = librosa.feature.mfcc(y=segment, sr=sr, n_mfcc=n_mfcc_coeffs).T

                        #n_frames = mfcc.shape[0]
                        pitch = librosa.yin(segment, fmin=75, fmax=600)
                        #intensity = librosa.feature.rms(y=y).flatten()
                        feature = np.concatenate((mfcc, np.expand_dims(pitch, axis=1)), axis=1)
                        if id in train_ids:
                            train_data['audio'].append(feature)
                            train_data['label'].append(i)
                            train_data['id'].append(id)
                            train_data['age'].append(age)
                            train_data['gender'].append(gender)
                        elif id in val_ids:
                            val_data['audio'].append(feature)
                            val_data['label'].append(i)
                            val_data['id'].append(id)
                            val_data['age'].append(age)
                            val_data['gender'].append(gender)
                        else:
                            print('Data {}/{} is not included.'.format(dir, file))

        for data in [train_data, val_data]:
            data['audio'] = torch.from_numpy(center_data(np.stack(data['audio']))).float()
            data['label'] = torch.tensor(data['label'])
            data['id'] = torch.tensor(data['id'])
            data['age'] = torch.tensor(data['age'])
            data['gender'] = torch.tensor(data['gender'])
        
        return train_data, val_data
    
    def load_n_seconds_audio_shift(self, seg_len, shift, train_ids, val_ids, n_mfcc_coeffs, prefix='Full_'):
        train_data = {'audio': [], 'label': [], 'id': [], 'age': [], 'gender': []}
        val_data = {'audio': [], 'label': [], 'id': [], 'age': [], 'gender': []}
        full_audio_dirs = [os.path.join(self.data_dir, label) for label in self.labels]
        print('--- Loading and segmenting {} second audios with shift {} seconds'.format(seg_len, shift))
        for i, dir in enumerate(full_audio_dirs):
            for file in os.listdir(dir):
                if file.endswith(".wav"):
                    y, sr = librosa.load(os.path.join(dir, file))
                    seg_samples = int(seg_len * sr)
                    shift_samples = int(shift * sr)
                    total_segs = (len(y) - seg_samples) // shift_samples 
                    id = int(file.split('-')[0])
                    gender_idx = 0 if file.split('-')[1] == 'm' else 1
                    gender_one_hot = torch.zeros(2)
                    gender_one_hot[gender_idx] = 1
                    age_idx = self.ages.index((file.split('-')[-1]).split('.')[0])
                    age_one_hot = torch.zeros(len(self.ages))
                    age_one_hot[age_idx] = 1
                    for seg_index in range(total_segs):
                        start = seg_index * shift_samples
                        end = start + seg_samples
                        segment = y[start:end]
                        
                        mfcc = librosa.feature.mfcc(y=segment, sr=sr, n_mfcc=n_mfcc_coeffs).T

                        #n_frames = mfcc.shape[0]
                        pitch = librosa.yin(segment, fmin=75, fmax=600)
                        #intensity = librosa.feature.rms(y=y).flatten()
                        feature = np.concatenate((mfcc, np.expand_dims(pitch, axis=1)), axis=1)
                        if id in train_ids:
                            train_data['audio'].append(feature)
                            train_data['label'].append(i)
                            train_data['id'].append(id)
                            train_data['age'].append(age_one_hot)
                            train_data['gender'].append(gender_one_hot)
                        elif id in val_ids:
                            val_data['audio'].append(feature)
                            val_data['label'].append(i)
                            val_data['id'].append(id)
                            val_data['age'].append(age_one_hot)
                            val_data['gender'].append(gender_one_hot)
                        else:
                            print('Data {}/{} is not included.'.format(dir, file))

        for data in [train_data, val_data]:
            data['audio'] = torch.from_numpy(center_data(np.stack(data['audio']))).float()
            data['label'] = torch.tensor(data['label'])
            data['id'] = torch.tensor(data['id'])
            data['age'] = torch.stack(data['age'], dim=0)
            data['gender'] = torch.stack(data['gender'], dim=0)
        
        return train_data, val_data




# Using pretrained model and cluster to do find data mistake
import torch
import torch.nn as nn

def find_mistakes(audio, label, id):
    vocoder = torch.hub.load('seungwonpark/melgan', 'melgan')
    vocoder.eval()
    if len(audio.shape) != 3:
        raise ValueError('--- Mel shape {} should be (n_samples, frames, n_mfcc_coeffs)'.format(audio.shape))
    
    if torch.cuda.is_available():
        vocoder = vocoder.cuda()
        audio = audio.cuda()
        label = label.cuda()
        id = id.cuda
        
    print(vocoder.mel_channel)
    with torch.no_grad():
        print(audio.shape)
        audio = vocoder.inference(audio)
        print(audio.shape)
        print(audio[0])
        print(audio)


if __name__ == "__main__":
    pass


import torch
import torch.nn as nn
import torch.nn.functional as F

class conv_mlp(nn.Module):
    def __init__(self, args):
        super(conv_mlp, self).__init__()
        self.args = args
        
        self.conv1 = nn.Conv2d(1, 16, kernel_size=(3, 3), stride=(1, 1))
        self.pool1 = nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2))
        self.conv2 = nn.Conv2d(16, 32, kernel_size=(3, 3), stride=(1, 1))
        self.pool2 = nn.MaxPool2d(kernel_size=(2, 2), stride=(2, 2))
        
        self.fc1 = nn.Linear(4960, 128)
        self.fc2 = nn.Linear(128, 64)
        self.fc3 = nn.Linear(64, 5)
        self.dropout = nn.Dropout(p=0.5)
        
        self.gender_embedding = nn.Linear(self.args.n_gender_classes, self.args.gender_embed_dim)
        self.age_embedding = nn.Linear(self.args.n_age_classes, self.args.age_embed_dim)

    def forward(self, x, gender, age):
        assert self.args.n_gender_classes == gender.shape[1]
        assert self.args.n_age_classes == age.shape[1]
        assert x.shape == (8, 171, 21), "{}".format(x.shape)
        x = self.conv1(x.unsqueeze(1))
        x = F.relu(x)
        x = self.pool1(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = self.pool2(x)
        x = x.view(8, -1, )
        
        gender_embed = self.gender_embedding(gender)
        age_embed = self.age_embedding(age)
        gender_embed = gender_embed.view(-1, self.args.gender_embed_dim)
        age_embed = age_embed.view(-1, self.args.age_embed_dim)
        
        #import pdb;pdb.set_trace()
        x = torch.cat((x, gender_embed, age_embed), dim=1)
        
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.fc3(x)
        return x


import torch
import torch.nn as nn

class MelGanClassify(nn.Module):
    def __init__(self, ) -> None:
        super(MelGanClassify).__init__()
        self.melgan = torch.hub.load('seungwonpark/melgan', 'melgan').cuda()
        self.fc = nn.Linear()
        
    def forward(self, mel):
        audio = self.melgan.inference(mel)
        return audio
        
    def pad1D(self, x, dim):
        pass


import os
import torch
import torch.nn as nn
import torchaudio
import torchaudio.functional as F
import torchaudio.transforms as T
import soundfile as sf
import re
import librosa
import pickle
import numpy as np
import random

from torch.utils.data import Dataset, DataLoader, random_split
from tqdm import tqdm

from utils import center_data, flatten_list

class SegmentedAudio(Dataset):
    def __init__(self, data):
        self.data = data
        self.len = data['audio'].shape[0]
        if len(data['audio'].shape) == 3:
            self.n_flatten_feats = data['audio'].shape[1] * data['audio'].shape[2]
        elif len(data['audio'].shape) == 2:
            self.n_flatten_feats = data['audio'].shape[-1]
        else:
            raise ValueError('Feature shape {} is not available.'.format(data['audio'].shape))
            
    def __len__(self):
        return self.len

    def __getitem__(self, idx):
        return self.data['audio'][idx], self.data['label'][idx], self.data['id'][idx], self.data['idx'][idx]
    
    def shape(self):
        return self.data['audio'].shape


class Mix:
    def __init__(self, args):
        self.data_dir = args.data_dir
        self.labels = args.labels
        self.seg_len = args.seg_len
        self.shift = args.shift
        self.n_mfcc_coeffs = args.n_mfcc_coeffs
        self.remove_ids = args.remove_ids
        self.seg_save_dir = args.seg_save_dir
        self.center_data = args.center_data
        if self.seg_save_dir is not None:
            if os.path.exists(self.seg_save_dir):
                raise ValueError('Directory of segmented audios exists!')
            else:
                for type in ['train', 'val', 'test']:
                    os.makedirs(os.path.join(self.seg_save_dir, type))
        self.audio_dirs = [os.path.join(self.data_dir, label) for label in self.labels]
        self.cls_num = [0] * len(self.labels)
        
        self.id2audios, self.n_baby, self.baby_ids = self.get_dataset_ids()
        args.train_ids, args.val_ids, args.test_ids = self.generate_train_val_test_ids(self.baby_ids, self.n_baby, args.split_type, args.load_path, args.save_path)
        print('--- Number of baby, train: {}, valid: {}, tests: {}'.format(len(args.train_ids), len(args.val_ids), len(args.test_ids)))
        if self.seg_len == 1:
            train_data, val_data = self.load_one_second_audio(args.train_ids, args.val_ids, args.n_mfcc_coeffs)
        elif self.seg_len == 0:
            train_data, val_data = self.load_full_audio(args.train_ids, args.val_ids, args.n_mfcc_coeffs)
        else:
            train_data, val_data, test_data = self.load_n_seconds_audio_shift(self.seg_len, self.shift, args.train_ids, args.val_ids, args.test_ids, args.n_mfcc_coeffs)
                
        if torch.cuda.is_available():
            for data in [train_data, val_data, test_data]:
                for k in data.keys():
                    data[k] = data[k].cuda()
        
        all_data = {}
        for k in train_data.keys():
            all_data[k] = torch.cat([train_data[k], val_data[k], test_data[k]], axis=0)

        self.all_data = all_data
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data
        
        self.train_dataset = SegmentedAudio(train_data)
        self.val_dataset = SegmentedAudio(val_data)
        self.test_dataset = SegmentedAudio(test_data)
        self.all_dataset = SegmentedAudio(all_data)
            
        print('--- Training samples: {}, validation samples: {}, validation samples: {}, total samples: {}'.format(len(self.train_dataset), \
            len(self.val_dataset), len(self.test_dataset), len(self.all_dataset)))
        print('--- Class number: {}'.format({label: num for label, num in zip(args.labels, self.cls_num)}))
        
    def get_dataset_ids(self):
        id2audios = {}
        n_baby = 0
        baby_ids = []

        for idx, dir in enumerate(self.audio_dirs):
            class_ids = []
            for file in os.listdir(dir):
                if file.lower().endswith(".wav"):
                    # record individual ids
                    infos = file.split('-')
                    id = int(infos[0])
                    if id not in self.remove_ids:
                        if id in id2audios:
                            id2audios[id].append(file)
                        else:
                            id2audios[id] = [file]
                            n_baby += 1
                            class_ids.append(id)

            if idx == 1: # for hungry dataset, we need more augmented data to improve its generalization
                n = len(class_ids) // 3
                new_class_ids = class_ids[:n]
                for i in [0,4,6,7,9,10,17,20,21,22]:
                    if i not in new_class_ids:
                        new_class_ids.append(i)
                baby_ids.append(new_class_ids)
            else:
                baby_ids.append(class_ids)
            
        print("--- Total baby number: {}, class baby num: {}".format(n_baby, [f"{self.labels[i]}: {len(cls)}" \
                for i, cls in enumerate(baby_ids)]))
        
        return id2audios, n_baby, baby_ids
        
    def generate_train_val_test_ids(self, individual_ids, n_baby, split_type, save_path, load_path):
        if load_path is not None:
            with open(load_path, "rb") as file:
                data = pickle.load(file)
            train_ids = data['train']
            val_ids = data['val']
            test_ids = data['test']
            assert len(val_ids + train_ids + test_ids) == len(flatten_list(individual_ids))
        else:
            if split_type == 'class':
                train_ids = []
                val_ids = []
                test_ids = []
                for cls in individual_ids:
                    random.shuffle(cls)
                    n_val = max(1, round(len(cls) * 0.2))
                    n_test = max(1, round(len(cls) * 0.2))
                    n_train = len(cls) - n_val - n_test
                    train_ids.extend(cls[:n_train])
                    val_ids.extend(cls[n_train:n_train + n_val])
                    test_ids.extend(cls[-n_test:])
                    
            elif split_type == 'baby':
                
                individual_ids = flatten_list(individual_ids)
                random.shuffle(individual_ids)
                n_train_individual = int(0.8 * n_baby)
                n_val_individuals = int(0.1 * n_baby)
                n_test_individuals = n_baby - n_train_individuals - n_val_individuals
                train_ids = individual_ids[:n_train_individuals]
                val_ids = individual_ids[n_train_individuals:n_train_individuals + n_val_individuals]
                test_ids = individual_ids[-n_test_individuals:]
            else:
                raise ValueError('Unknow splitting type {}'.format(split_type))
            
            if save_path is not None:
                data = {'train': train_ids, 'val': val_ids}
                with open(save_path, "wb") as file:
                    pickle.dump(data, file)
        
        return train_ids, val_ids, test_ids

    def load_n_seconds_audio_shift(self, seg_len, shift, train_ids, val_ids, test_ids, n_mfcc_coeffs, prefix='Full_'):
        train_data = {'audio': [], 'label': [], 'id': [], 'idx': []}
        val_data = {'audio': [], 'label': [], 'id': [], 'idx': []}
        test_data = {'audio': [], 'label': [], 'id': [], 'idx': []}
        print('--- Loading and segmenting {} second audios with shift {}'.format(seg_len, shift))
        for i, dir in enumerate(self.audio_dirs):
            for file in os.listdir(dir):
                if file.endswith(".wav"):
                    
                    infos = file[:-4].split('-')
                    id = int(infos[0])
                    idx = int(infos[1])
                    gender = infos[2]
                    age = infos[3]
                    
                    y, sr = librosa.load(os.path.join(dir, file))
                    # Add white noise
                    # if id in train_ids:
                    #     noise_factor = 0.0001
                    #     white_noise = np.random.randn(len(y)) * noise_factor
                    #     y = y + white_noise
                    
                    seg_samples = int(seg_len * sr)
                    shift_samples = int(shift * sr)
                    total_segs = max((len(y) - seg_samples) // shift_samples + 1, 1)
                    for seg_index in range(total_segs):
                        start = seg_index * shift_samples
                        end = start + seg_samples
                        if end - start > len(y):
                            segment = np.pad(y, (0, end - start - len(y)), mode='constant')
                        else:
                            segment = y[start:end]
                        
                        mfcc = librosa.feature.mfcc(y=segment, sr=sr, n_mfcc=n_mfcc_coeffs).T
                        
                        #n_frames = mfcc.shape[0]
                        pitch = librosa.yin(segment, fmin=75, fmax=600)
                        #intensity = librosa.feature.rms(y=y).flatten()
                        feature = np.concatenate((mfcc, np.expand_dims(pitch, axis=1)), axis=1)
                        
                        if id in train_ids or id in self.remove_ids:
                            type = 'train'
                            train_data['audio'].append(feature)
                            train_data['label'].append(i)
                            train_data['id'].append(id)
                            train_data['idx'].append(idx)
                        elif id in val_ids:
                            type = 'val'
                            val_data['audio'].append(feature)
                            val_data['label'].append(i)
                            val_data['id'].append(id)
                            val_data['idx'].append(idx)
                        elif id in test_ids:
                            type = 'test'
                            test_data['audio'].append(feature)
                            test_data['label'].append(i)
                            test_data['id'].append(id)
                            test_data['idx'].append(idx)

                        self.cls_num[i] += 1
                        
                        if self.seg_save_dir is not None:
                            seg_save_path = os.path.join(self.seg_save_dir, type, '{}-{}-{}-{}-{}.wav'.format(id, idx, seg_index, gender, age))
                            #librosa.output.write_wav(seg_save_path, segment, sr) libsora.output has been removed in 0.8.0
                            sf.write(seg_save_path, segment, sr)

        for data in [train_data, val_data, test_data]:
            np_audios = np.stack(data['audio'])
            if self.center_data:
                np_audios = center_data(np_audios)
            data['audio'] = torch.from_numpy(np_audios).float() 
            data['label'] = torch.tensor(data['label'])
            data['id'] = torch.tensor(data['id'])
            data['idx'] = torch.tensor(data['idx'])
        
        return train_data, val_data, test_data




import torch.nn as nn

class cross_entropy_loss(nn.Module): 
    def __init__(self, args):
        super(cross_entropy_loss, self).__init__()
        if args.reweight:
            self.criterion = nn.CrossEntropyLoss(weight=args.cls_weights)
        else:
            self.criterion = nn.CrossEntropyLoss()
        self.args = args
        
    def forward(self, projections, targets):
        loss = self.criterion(projections, targets)
        return loss


import torch
import torch.nn as nn
import torch.nn.functional as F

class mlp(nn.Module):
    def __init__(self, in_feats, mid_feats, out_feats, n_classes, dropout=0.5):
        super(mlp, self).__init__()
    
        self.fc1 = nn.Linear(in_feats, mid_feats)
        self.bn1 = nn.BatchNorm1d(mid_feats)
        self.fc2 = nn.Linear(mid_feats, out_feats)
        self.bn2 = nn.BatchNorm1d(out_feats)
        self.fc3 = nn.Linear(out_feats, n_classes)
        self.relu = nn.ReLU(inplace=True)
        self.fc4 = nn.Linear(out_feats, out_feats)
        self.bn3 = nn.BatchNorm1d(out_feats)
        self.fc5 = nn.Linear(out_feats, out_feats)

        if dropout is not None:
            self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        if hasattr(self, 'dropout'):
            x1 = self.dropout(self.relu(self.bn1(self.fc1(x))))
        else:
            x1 = self.relu(self.bn1(self.fc1(x)))
            
        if hasattr(self, 'dropout'):
            x2 = self.dropout(self.relu(self.bn2(self.fc2(x1))))
        else:
            x2 = self.relu(self.fc2(x1))
            
        x3 = self.fc3(x2)
        
        return x3

    def forward_contrastive(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        if hasattr(self, 'dropout'):
            x1 = self.dropout(self.relu(self.bn1(self.fc1(x))))
        else:
            x1 = self.relu(self.bn1(self.fc1(x)))
            
        if hasattr(self, 'dropout'):
            x2 = self.dropout(self.relu(self.bn2(self.fc2(x1))))
        else:
            x2 = self.relu(self.fc2(x1))
            

        x2 =  self.fc4(x2)
        x2 = F.relu(x2)
        x2 = self.fc5(x2)
        # Normalize to unit hypersphere
        x2 = F.normalize(x2, dim=1)

        return x2

class mlpv2(nn.Module):
    def __init__(self, in_feats, mid_feats, out_feats, n_classes, dropout=0.5):
        super(mlp, self).__init__()
        self.fc1 = nn.Linear(in_feats, mid_feats)
        self.bn1 = nn.BatchNorm1d(mid_feats)
        self.fc2 = nn.Linear(mid_feats, out_feats)
        self.bn2 = nn.BatchNorm1d(out_feats)
        self.fc3 = nn.Linear(out_feats, n_classes)
        self.relu = nn.ReLU(inplace=True)
        self.fc4 = nn.Linear(out_feats, out_feats)
        self.bn3 = nn.BatchNorm1d(out_feats)
        self.fc5 = nn.Linear(out_feats, out_feats)

        if dropout is not None:
            self.dropout = nn.Dropout(p=dropout)

    def forward(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        if hasattr(self, 'dropout'):
            x1 = self.dropout(self.relu(self.bn1(self.fc1(x))))
        else:
            x1 = self.relu(self.bn1(self.fc1(x)))
            
        if hasattr(self, 'dropout'):
            x2 = self.dropout(self.relu(self.bn2(self.fc2(x1))))
        else:
            x2 = self.relu(self.fc2(x1))
            
        x3 = self.fc3(x2)
        
        return x3

    def forward_contrastive(self, x):
        if len(x.shape) > 2:
            x = torch.flatten(x, start_dim=1, end_dim=len(x.shape) - 1)
        if hasattr(self, 'dropout'):
            x1 = self.dropout(self.relu(self.bn1(self.fc1(x))))
        else:
            x1 = self.relu(self.bn1(self.fc1(x)))
            
        if hasattr(self, 'dropout'):
            x2 = self.dropout(self.relu(self.bn2(self.fc2(x1))))
        else:
            x2 = self.relu(self.fc2(x1))
            

        x2 =  self.fc4(x2)
        x2 = F.relu(x2)
        x2 = self.fc5(x2)
        # Normalize to unit hypersphere
        x2 = F.normalize(x2, dim=1)

        return x2


import librosa
import numpy as np
import argparse
import torch
import librosa
import numpy as np
import noisereduce as nr
import glob
import os

from model import resnet
from collections import Counter

def most_frequent(list_of_elements):
    """
    Returns the most frequent element in the given list.
    
    Parameters:
    - list_of_elements: A array-like list containing the elements to analyze.
    
    Returns:
    - The most frequent element in the list.
    """
    counter = Counter(list_of_elements)
    most_common = counter.most_common(1)[0][0]
    return most_common

def extract_most_energetic_segments(y, sr, segment_length=2.0, n_segments=1):
    """
    Extracts the top-n {segment_length}-seconds segments with the highest energy from an audio file.

    Parameters:
    - y: Audio time series.
    - sr: Sampling rate of y.
    - segment_length: Length of the segment to extract in seconds.
    - n_segments: Number of energetic segments to return.

    Returns:
    - List of tuples containing start and end times and the audio data of the most energetic segments.
    """
    segment_samples = int(segment_length * sr)
    segments_energy = []

    # Calculate energy for each segment and store with start sample index
    for start_sample in range(0, len(y) - segment_samples, segment_samples // 2):  # 50% overlap
        segment = y[start_sample:start_sample + segment_samples]
        energy = np.sum(segment**2)
        segments_energy.append((start_sample, energy))

    # Sort segments by energy in descending order and select top n_segments
    top_segments = sorted(segments_energy, key=lambda x: x[1], reverse=True)[:n_segments]

    # Extract the audio data for the top n_segments
    energetic_segments = []
    for start_sample, _ in top_segments:
        start_time = start_sample / sr
        end_time = (start_sample + segment_samples) / sr
        segment_data = y[start_sample:start_sample + segment_samples]
        energetic_segments.append((start_time, end_time, segment_data))

    return energetic_segments

def useful_information_proportion(y, sr, silence_threshold=0.01, frame_length=2048, hop_length=512):
    """
    Calculate the proportion of 'useful information' in an audio file based on a simple silence detection algorithm.

    Parameters:
    - audio_path: Path to the audio file.
    - silence_threshold: The threshold value below which sound is considered silence.
    - frame_length: The number of samples per analysis frame.
    - hop_length: The number of samples to shift the analysis window for each frame.

    Returns:
    - Proportion of the audio file that is above the silence threshold.
    """

    energy = np.array([sum(abs(y[i:i+frame_length]**2))
                       for i in range(0, len(y), hop_length)])

    normalized_energy = energy / np.max(energy)

    non_silent_frames = np.sum(normalized_energy > silence_threshold)

    total_frames = len(energy)
    proportion_non_silent = non_silent_frames / total_frames

    return proportion_non_silent

def inference(audio_path, model, checkpoint_path, seg_len=2, n_classes=4, n_segments=3, n_mfcc_coeffs=20):
    """
    Infers the label of an audio file using a trained model (resnet).
    Parameters:
    - audio_path: Path to the audio file.
    - model: torch.nn.Module object. The trained model to use for inference.
    - args: argparse.Namespace object. The command-line arguments of the script.
    
    Returns:
    - start_time, end_time: Start and end times of the most energetic segment in the audio file.
    - energetic_segment: The audio data of the most energetic segment.
    """
    y, sr = librosa.load(audio_path)
    y = nr.reduce_noise(y=y, sr=sr)
    
    score = useful_information_proportion(y, sr)
    segs = extract_most_energetic_segments(y, sr, seg_len, n_segments)
    model = resnet(n_classes)
    if torch.cuda.is_available():
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')
        
    model.load_state_dict(torch.load(checkpoint_path, map_location=device))
    model = model.to(device)
    model.eval()
    feature_tensors = []
    for start_time, end_time, seg in segs:
        mfcc = librosa.feature.mfcc(y=seg, sr=sr, n_mfcc=n_mfcc_coeffs).T
        pitch = librosa.yin(seg, fmin=75, fmax=600)
        feature = np.concatenate((mfcc, np.expand_dims(pitch, axis=1)), axis=1)
        feature_tensor = torch.from_numpy(feature).unsqueeze(0).float().to(device)
        feature_tensors.append(feature_tensor)
        
    batched_tensor = torch.cat(feature_tensors, dim=0)
    
    output = model(batched_tensor)
    print(f'Model Output: \n {output}')
    probs = torch.nn.functional.softmax(output, dim=-1)
    print(f'Probability: \n {probs}')
    predictions = torch.argmax(probs, dim=-1)
        
    predicted_label = most_frequent(predictions.cpu().numpy())
    return predicted_label, score


    
if __name__ == "__main__":
    N_CLASSES = 4
    AUDIO_PATH = './files/audio/'
    CHECKPOINT_PATH = './files/checkpoint/resnet_1.pth'
    SEG_LEN = 2.0
    N_SEGMENTS = 3
    N_MFCC_COEFFS = 20
    
    paths = glob.glob(os.path.join(AUDIO_PATH, '*.wav'))
    
    for path in paths:
        model = resnet(N_CLASSES)
        predicted_label, score = inference(path, model, CHECKPOINT_PATH, seg_len=SEG_LEN, n_segments=N_SEGMENTS, n_mfcc_coeffs=N_MFCC_COEFFS)
        print(f'Most Frequent Predicted Label: {predicted_label}, Useful Score {score}')


import torch
import torch.nn as nn

class AudioClassifier(nn.Module):
    def __init__(self, input_dim):
        super(AudioClassifier, self).__init__()
        self.fc1 = nn.Linear(input_dim, 128)
        self.fc2 = nn.Linear(128, 64)
        self.fc3 = nn.Linear(64, 2)
        self.relu = nn.ReLU()
    
    def forward(self, x):
        x = torch.flatten(x, start_dim=1)
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        return self.fc3(x)


from setuptools import setup, find_packages
from setuptools.command.install import install
import getpass
import sys
import os

class CustomInstallCommand(install):
    def run(self):
        password = getpass.getpass('Enter the installation password: ')
        if password == 'cradleai.':
            install.run(self)
        else:
            print("Invalid password")
            sys.exit(1)

# Replace with your project name
project_name = "letbabytalk"

# Replace with your project description
project_description = "letbabytalk"

# Replace with your name
author = "Minghao Fu"

# Replace with your email
author_email = "isminghaofu@gmail.com"

# Replace with the license you're using (e.g. MIT, Apache 2.0)
license = "MIT"

# Add any requirements your project needs (e.g. ["numpy", "tensorflow"])
requirements = []

setup(
    name=project_name,
    version="0.1.4",  # Ensure versioning is consistent with your release strategy
    description=project_description,
    long_description=open("README.md", "r", encoding="utf-8").read() if os.path.exists("README.md") else "",
    long_description_content_type="text/markdown" if os.path.exists("README.md") else None,
    author=author,
    author_email=author_email,
    license=license,
    install_requires=requirements,  # Ensure your requirements are a list of dependencies
    packages=find_packages(),  # Automatically finds packages in your project
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        # Add more classifiers if necessary
    ],
    python_requires='>=3.6',  # Ensure this matches the minimum Python version your package supports
    cmdclass={
        'install': CustomInstallCommand,  # Uncomment if you have a custom install command
    },
)



import torch
import torch.nn as nn

class supervised_contrastive_loss(nn.Module):
    def __init__(self, args):
        """
        Implementation of the loss described in the paper Supervised Contrastive Learning :
        https://arxiv.org/abs/2004.11362

        :param temperature: int
        """
        super(supervised_contrastive_loss, self).__init__()
        self.temperature = args.temperature
        print(self.temperature)

    def forward(self, projections, targets):
        """
        :param projections: torch.Tensor, shape [batch_size, projection_dim]
        :param targets: torch.Tensor, shape [batch_size]
        :return: torch.Tensor, scalar
        """
        device = torch.device("cuda") if projections.is_cuda else torch.device("cpu")

        dot_product_tempered = torch.mm(projections, projections.T) / self.temperature
        # Minus max for numerical stability with exponential. Same done in cross entropy. Epsilon added to avoid log(0)
        exp_dot_tempered = (
            torch.exp(dot_product_tempered - torch.max(dot_product_tempered, dim=1, keepdim=True)[0]) + 1e-5
        )

        mask_similar_class = (targets.unsqueeze(1).repeat(1, targets.shape[0]) == targets).to(device)
        mask_anchor_out = (1 - torch.eye(exp_dot_tempered.shape[0])).to(device)
        mask_combined = mask_similar_class * mask_anchor_out
        cardinality_per_samples = torch.sum(mask_combined, dim=1)

        log_prob = -torch.log(exp_dot_tempered / (torch.sum(exp_dot_tempered * mask_anchor_out, dim=1, keepdim=True)))
        supervised_contrastive_loss_per_sample = torch.sum(log_prob * mask_combined, dim=1) / (cardinality_per_samples + 1e-5)
        supervised_contrastive_loss = torch.mean(supervised_contrastive_loss_per_sample)

        return supervised_contrastive_loss


import os
import librosa
import numpy as np
import random
import glob
import argparse
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision.models as models
from torch.utils.data import Dataset, DataLoader
from sklearn.model_selection import train_test_split
from tqdm import tqdm
from Caulimate.Utils.Tools import seed_everything, makedir
import sys
sys.path.append('..')
from models.resnet import bin_resnet  
seed_everything(42)

DATASET_DIR = '/l/users/minghao.fu/minghao.fu/dataset'
LOG_DIR = '/l/users/minghao.fu/minghao.fu/logs'

POS_DIR_PATH = os.path.join(DATASET_DIR, 'AudioSet/unbalanced_train/Baby cry, infant cry')
NEG_DIR_PATH =  os.path.join(DATASET_DIR, 'AudioSet/balanced_train')
SAVE_DIR_PATH = os.path.join('./files/checkpoint')

N_MFCC_COEFFS = 13
WINDOW_SIZE = 100
THRESHOLD = 0.05
BATCH_SIZE = 32
LEARNING_RATE = 0.001
NUM_EPOCHS = 200

class AudioDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels
    
    def __len__(self):
        return len(self.features)
    
    def __getitem__(self, idx):
        feature = torch.tensor(self.features[idx], dtype=torch.float32)
        label = torch.tensor(self.labels[idx], dtype=torch.long)
        return feature, label

class AudioClassifier(nn.Module):
    def __init__(self, num_classes=2):
        super(AudioClassifier, self).__init__()
        self.resnet = models.resnet18(pretrained=True)
        num_ftrs = self.resnet.fc.in_features
        self.resnet.fc = nn.Linear(num_ftrs, num_classes)

    def forward(self, x):
        return self.resnet(x)
    
def get_samples(directory_path):
    samples = []
    for subdir, _, files in os.walk(directory_path):
        for file in files:
            if file.endswith('.wav'):
                file_path = os.path.join(subdir, file)
                samples.append((file_path, os.path.basename(subdir)))
    return samples

def extract_features(file_path, n_mfcc_coeffs=N_MFCC_COEFFS, threshold=THRESHOLD):
    segments = []
    y, sr = librosa.load(file_path)
    # Calculate the number of frames that correspond to 2 seconds
    window_size = 2 * sr  # adjust this if you're using a different hop length
    for seg_idx, start in enumerate(range(0, len(y) - window_size + 1, window_size // 2)):
        end = start + window_size
        if end >= y.shape[0]:
            break
        seg_y = y[start:end]
        mfcc = librosa.feature.mfcc(y=seg_y, sr=sr, n_mfcc=n_mfcc_coeffs).T
        pitch = np.expand_dims(librosa.yin(seg_y, fmin=75, fmax=600), axis=1)
        feature = np.concatenate((mfcc, pitch), axis=1)
        segments.append(feature)

    return segments

def train():

    makedir(SAVE_DIR_PATH)
    negative_samples = random.sample(get_samples(NEG_DIR_PATH), 1000)
    positive_samples = random.sample(get_samples(POS_DIR_PATH), 1000)
    

    neg_features = [seg for fp, _ in tqdm(negative_samples) for seg in extract_features(fp)]

    neg_labels = [0] * len(neg_features)
   

    pos_features = [seg for fp, _ in tqdm(positive_samples) for seg in extract_features(fp)]
    pos_labels = [1] * len(pos_features)
    for each in pos_features:
        if each.shape != pos_features[0].shape:
            import pdb; pdb.set_trace()

    features = np.array(neg_features + pos_features)
    labels = np.array(neg_labels + pos_labels)

    X_train, X_val_test, y_train, y_val_test = train_test_split(features, labels, test_size=0.2, random_state=42)

    X_test, X_val, y_test, y_val = train_test_split(X_val_test, y_val_test, test_size=0.5, random_state=42)


    train_loader = DataLoader(AudioDataset(X_train, y_train), batch_size=BATCH_SIZE, shuffle=True)
    val_loader = DataLoader(AudioDataset(X_val, y_val), batch_size=BATCH_SIZE)
    test_loader = DataLoader(AudioDataset(X_test, y_test), batch_size=BATCH_SIZE)

    model = bin_resnet()
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=LEARNING_RATE)

    min_val_loss = float('inf')
    for epoch in range(NUM_EPOCHS):
        # save_epoch_path = os.path.join(SAVE_DIR_PATH, f'epoch_{epoch}') 
        # makedir(save_epoch_path)    
        model.train()
        running_loss = 0.0
        for inputs, labels in train_loader:
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
            running_loss += loss.item()
        with torch.no_grad():
            val_loss = 0.0
            for inputs, labels in val_loader:
                outputs = model(inputs)
                loss = criterion(outputs, labels)
                val_loss += loss.item()
            if val_loss < min_val_loss:
                min_val_loss = val_loss
                print(f"{epoch} epoch, val_loss: {val_loss}, saving model to")
                torch.save(model.state_dict(), os.path.join(SAVE_DIR_PATH, f'e{epoch}_best_model.pth'))
                early_stop = 0
            else:
                early_stop += 1
            if early_stop == 10:
                break
        print(f'Epoch {epoch + 1}/{NUM_EPOCHS}, Loss: {running_loss / len(train_loader):.4f}')
        #torch.save(model.state_dict(), os.path.join(save_epoch_path, 'checkpoint.pth'))

    model.eval()
    correct = [0, 0]
    total = [0, 0]

    with torch.no_grad():
        for inputs, labels in test_loader:
            outputs = model(inputs)
            _, predicted = torch.max(outputs.data, 1)
            for label, prediction in zip(labels, predicted):
                total[label] += 1
                if label == prediction:
                    correct[label] += 1

    accuracy_0 = correct[0] / total[0] if total[0] > 0 else 0
    accuracy_1 = correct[1] / total[1] if total[1] > 0 else 0

    print(f'Accuracy for class 0: {accuracy_0 * 100:.2f}%')
    print(f'Accuracy for class 1: {accuracy_1 * 100:.2f}%')
    total_accuracy = sum(correct) / sum(total) if sum(total) > 0 else 0
    print(f'Total accuracy: {total_accuracy * 100:.2f}%')

def test(audio_dir='./files/audio/', ckp_path='./files/checkpoint/e17_best_model.pth'):
    audio_names = os.listdir(audio_dir)
    audio_paths = [os.path.join(audio_dir, name) for name in audio_names]
    model = bin_resnet()
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    state_dict = torch.load(ckp_path, map_location=device)

    # Load the state dictionary into the model
    model.load_state_dict(state_dict)
    model.eval()
    for audio_path in audio_paths:
        seg_predicts = []
        segments = extract_features(audio_path)
        for segment in segments:
            feature_tensor = torch.from_numpy(segment).unsqueeze(0).float()
            outputs = model(feature_tensor)
            _, predicted = torch.max(outputs.data, 1)
            seg_predicts.append(predicted.item())

        vote = max(seg_predicts, key=seg_predicts.count)
        print(f'Segment predictions: {seg_predicts}, vote result: {vote}')

def main():
    parser = argparse.ArgumentParser(description='Train or test the model')
    parser.add_argument('--mode', type=str, choices=['train', 'test'], default='train',
                        help='Mode to run: "train" or "test"')

    args = parser.parse_args()

    if args.mode == 'train':
        train()
    elif args.mode == 'test':
        test()  # Assume you have a function called test

if __name__ == '__main__':
    main()
